import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;
import java.util.Properties;


public class MoziloBackup
{

	public static void main(String[] args) {
		try {
			new MoziloBackup();
		}
		catch (Exception e) {
			System.err.println("Schwerer Fehler: "+ e.getMessage());
			e.printStackTrace();
		}
	}

	public MoziloBackup() throws Exception {
		String currentdate = getCurrentDateAsDirName();
        File logDir = new File("Logs");
        if (!logDir.exists())
        	logDir.mkdirs();
		File logFile = new File(logDir, currentdate+".log");
		FileWriter fw = new FileWriter(logFile); 
		BufferedWriter bw = new BufferedWriter(fw); 
		toScreenAndLog((new Date()).toString(), bw, true);
		toScreenAndLog("", bw, true);
		toScreenAndLog("########################", bw, true);
		toScreenAndLog("#                      #", bw, true);
		toScreenAndLog("#   moziloBackup 0.3   #", bw, true);
		toScreenAndLog("#                      #", bw, true);
		toScreenAndLog("########################", bw, true);
		toScreenAndLog("", bw, true);
		// Einstellungen laden
		toScreenAndLog("Lade Einstellungen aus data.txt...", bw, false);
		File file = new File("./data.txt");
        if (!file.exists()) {
        	toScreenAndLog("", bw, true);
    		toScreenAndLog("Die Einstellungsdatei data.txt konnte nicht gefunden werden! Programmende.", bw, true);
        	bw.close();
            System.exit(0);
        }
        toScreenAndLog("fertig.", bw, true);
        
       	bw.write("\r\n");
        String propFTP = getConfProperty(file, "server");
       	bw.write("               Server: " + propFTP + "\r\n");
       	String propPort = getConfProperty(file, "port");
       	bw.write("                 Port: " + propPort + "\r\n");
       	String propUser = getConfProperty(file, "user");
       	bw.write("             Username: " + propUser + "\r\n");
       	String propPass = getConfProperty(file, "password");
       	String propRemotedir = getConfProperty(file, "remotedir");
       	bw.write("     Entferntes Verz.: " + propRemotedir+ "\r\n");
       	String[] propRemoteexcludedirs = getConfProperty(file, "excludedirs").split(",");
       	bw.write("Ignorierte Unterverz.: " + getConfProperty(file, "excludedirs")+ "\r\n");
        String propProxy = getConfProperty(file, "proxyhost");
       	bw.write("                Proxy: " + propProxy + "\r\n");
        String propProxyport = getConfProperty(file, "proxyport");
       	bw.write("           Proxy-Port: " + propProxyport + "\r\n");
       	bw.write("\r\n");
       	
       	boolean propReplaceBackup = false;
        if (getConfProperty(file, "replacebackup").equals("1"))
        	propReplaceBackup = true;

        // Zum FTP verbinden
       	FTP ftp;
       	if (!propProxy.equals("") && !propProxyport.equals("")){
           	toScreenAndLog("Verbinde ueber Proxy " + propProxy + ":" + propProxyport + " mit " + propFTP + ":" + propPort + "...", bw, false);
       		ftp = new FTP(propFTP, Integer.parseInt(propPort), propProxy, Integer.parseInt(propProxyport));
       	}
       	else {
           	toScreenAndLog("Verbinde mit " + propFTP + ":" + propPort + "...", bw, false);
       		ftp = new FTP(propFTP, Integer.parseInt(propPort));
       	}
        getConnection(ftp, bw, propUser, propPass);

        // Backupverzeichnis erstellen, wenn noch nicht vorhanden
        File backupDir = new File("Backups");
        if (!backupDir.exists())
        	backupDir.mkdirs();
        // wenn propReplaceBackup == true, alle alten Backups vorher lschen 
        if (propReplaceBackup) {
        	toScreenAndLog("Leere lokales Verzeichnis " + backupDir.getName() + "...", bw, false);
        	deleteDir(backupDir);
        	backupDir.mkdirs();
            toScreenAndLog("fertig.", bw, true);
        }
        // Unterverzeichnis fr aktuelles Backup anlegen 
        File currentBackupDir = new File(backupDir, currentdate);
    	toScreenAndLog("Lege lokales Backupverzeichnis " + currentBackupDir.getName() + " an...", bw, false);
        currentBackupDir.mkdirs();
        toScreenAndLog("fertig.", bw, true);
        // Alles downloaden
        toScreenAndLog("", bw, true);
		long starttime = (new Date()).getTime();
        long size = downloadDir(ftp, currentBackupDir.getAbsolutePath(), propRemotedir, 0, bw, propUser, propPass, propRemoteexcludedirs);
        long endtime = (new Date()).getTime();
        double duration = endtime-starttime;
        duration = (duration*10)/10000;

        if (size != -1) {
        	toScreenAndLog("", bw, true);
        	toScreenAndLog("Backup erfolgreich beendet.", bw, true);
        	toScreenAndLog("  Benoetigte Zeit: " + duration + " s", bw, true);
        	toScreenAndLog("  Downloadvolumen: " + Math.ceil(size*100/(1024*1024))/100 + " MB (" + size + " B)", bw, true);
        	toScreenAndLog("Uebertragungsrate: " + Math.ceil(((size/duration)/1024)*1000)/1000 + " KB/s", bw, true);
        }
        else {
        	toScreenAndLog("", bw, true);
        	toScreenAndLog("Backup fehlerhaft!", bw, true);
        }
        toScreenAndLog("Trenne Verbindung...", bw, false);
		ftp.disconnect();
		toScreenAndLog("fertig", bw, true);
		toScreenAndLog("", bw, true);
		toScreenAndLog("", bw, true);
		toScreenAndLog("Fragen oder Anregungen? mozilo im Netz: http://www.mozilo.de", bw, true);
		toScreenAndLog("", bw, true);
		toScreenAndLog("", bw, true);
		bw.close();
    }
	
	public long downloadDir(FTP ftp, String localdir, String remotedir, long size, BufferedWriter bw, String propUser, String propPass, String[] propRemoteexcludedirs) throws Exception {
		ftp.cd(remotedir);
		if (ftp.lastCode() != FTP.CODE_CD_OK) {
			toScreenAndLog("Fehler: Verzeichnis " + remotedir + " existiert nicht!", bw, true);
			return -1;
		}
		ftp.list();
		RemoteFile[] remotefiles = ftp.getDirectoryList();
		for (int i=0; i<remotefiles.length; i++) {
			boolean excluded = false;
			// . und .. berspringen
			if (remotefiles[i].name.equals(".") || remotefiles[i].name.equals(".."))
				excluded = true;
			// ausgeschlossene Verzeichnisse berspringen
			for (int j=0; j<propRemoteexcludedirs.length; j++) {
		        if (remotefiles[i].name.equals(propRemoteexcludedirs[j]))
		        	excluded = true;
			}
			if (excluded)
				continue;
			if (remotefiles[i].isDirectory) {
				File tempfile = new File(localdir, remotefiles[i].name);
				tempfile.mkdir();
				long subsize = downloadDir(ftp, (new File(localdir, remotefiles[i].name)).getAbsolutePath(), remotefiles[i].name, 0, bw, propUser, propPass, propRemoteexcludedirs);
				if (subsize == -1)
					return -1;
				else
					size += subsize;
				ftp.cd("..");
			}
			else {
				File tempfile = new File(localdir, remotefiles[i].name);
				toScreenAndLog("Lade herunter: "+remotefiles[i].name + " (" +remotefiles[i].size + " B)...", bw, false);
				ftp.download(remotefiles[i].name, tempfile.getPath());
				// bei Fehler: max. 19 Wiederholungen
				boolean error = false;
				for(int j=2; j<=19; j++) {
					if (ftp.lastCode() == FTP.CODE_TRANSFER_OK) {
						error = false;
						continue;
					}
					else {
						error = true;
						toScreenAndLog("Transferfehler. ", bw, true);
						bw.write(ftp.lastCode() + ": " + getMessageByNumber(ftp.lastCode()) + "\r\n");
						// Bei mehr als fnf Fehlversuchen: erneuter Login 
						if (j>=5) {
							ftp.disconnect();
							getConnection(ftp, bw, propUser, propPass);
						}
						toScreenAndLog("Versuch " + j + "...", bw, false);
						ftp.download(remotefiles[i].name, tempfile.getPath());
					}
				}
				if (error) {
					toScreenAndLog("Wiederholter Uebertragungsfehler. Abbruch.", bw, true);
					return -1;
				}
				else {
					toScreenAndLog("fertig.", bw, true);
					bw.write(ftp.lastCode() + ": " + getMessageByNumber(ftp.lastCode()) + "\r\n");
					size += remotefiles[i].size;
				}
			}
		}
		return size;
	}

    public String getConfProperty(File file, String property) {
        Properties prop = null;
        String value = "";
        try
        {
            FileInputStream inputStream = new FileInputStream(file);

            prop = new Properties();
            prop.load(inputStream);
            value = prop.getProperty(property);
            inputStream.close();
        }
        catch (FileNotFoundException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        catch (IOException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        if (value == null)
        	return "";
        else
        	return value;
    }

    public String getCurrentDateAsDirName() {
        Date date = new Date();
        String month = Integer.toString(date.getMonth()+1);
        if (month.length() == 1)
        	month = "0"+month;
        String day = Integer.toString(date.getDate());
        if (day.length() == 1)
        	day = "0"+day;
        String hour = Integer.toString(date.getHours());
        if (hour.length() == 1)
        	hour = "0"+hour;
        String minute = Integer.toString(date.getMinutes());
        if (minute.length() == 1)
        	minute = "0"+minute;
        String second = Integer.toString(date.getSeconds());
        if (second.length() == 1)
        	second = "0"+second;
        return Integer.toString(date.getYear()+1900) + "-"
        		+ month + "-"
				+ day + "_"
				+ hour + "-"
				+ minute + "-"
				+ second;
    }
    
    public String getMessageByNumber(int number) {
    	switch(number) {
    		case 200: return "Command okay.";
    		case 202: return "Command not implemented, superfluous at this site.";
    		case 211: return "System status, or system help reply.";
    		case 212: return "Directory status.";
    		case 213: return "File status.";
    		case 214: return "Help message. On how to use the server or the meaning of a particular non-standard command. This reply is useful only to the human user.";
    		case 215: return "NAME system type. Where NAME is an official system name from the list in the Assigned Numbers document.";
    		case 220: return "Service ready for new user.";
    		case 221: return "Service closing control connection.";
    		case 225: return "Data connection open; no transfer in progress.";
    		case 226: return "Closing data connection. Requested file action successful.";
    		case 227: return "Entering Passive Mode (h1,h2,h3,h4,p1,p2).";
    		case 230: return "User logged in, proceed. Logged out if appropriate.";
    		case 250: return "Requested file action okay, completed.";
    		case 257: return "PATHNAME created.";
    		case 331: return "User name okay, need password.";
    		case 332: return "Need account for login.";
    		case 350: return "Requested file action pending further information";
    		case 421: return "Service not available, closing control connection.This may be a reply to any command if the service knows it must shut down.";
    		case 425: return "Can't open data connection.";
    		case 426: return "Connection closed; transfer aborted.";
    		case 450: return "Requested file action not taken.";
    		case 451: return "Requested action aborted. Local error in processing.";
    		case 452: return "Requested action not taken. Insufficient storage space in system.File unavailable (e.g., file busy).";
    		case 500: return "Syntax error, command unrecognized. This may include errors such as command line too long.";
			case 501: return "Syntax error in parameters or arguments.";
			case 502: return "Command not implemented.";
			case 503: return "Bad sequence of commands.";
			case 504: return "Command not implemented for that parameter.";
			case 530: return "Not logged in.";
			case 532: return "Need account for storing files.";
			case 550: return "Requested action not taken. File unavailable (e.g., file not found, no access).";
			case 551: return "Requested action aborted. Page type unknown.";
			case 552: return "Requested file action aborted. Exceeded storage allocation (for current directory or dataset).";
			case 553: return "Requested action not taken. File name not allowed.";
    		default:  return "Unknown Message!";
    	}
    }
    
    public void toScreenAndLog(String msg, BufferedWriter bw, boolean newline) throws Exception {
    	System.out.print(msg);
    	bw.write(msg);
    	if (newline) {
    		System.out.println();
    		bw.write("\r\n");
    	}
    }
    
    public void getConnection(FTP ftp, BufferedWriter bw, String propUser, String propPass) throws Exception {
        ftp.connect();
        if (ftp.lastCode() != FTP.CODE_CONNECT_OK) {
        	toScreenAndLog("", bw, true);
        	toScreenAndLog("Fehler: Die Verbindung zum FTP-Server konnte nicht aufgebaut werden. Programmende.", bw, true);
    		bw.close();
            System.exit(0);
        }
        ftp.setTransfer(FTP.TRANSFER_PASV);
        ftp.setMode(FTP.MODE_AUTO);
        toScreenAndLog("fertig.", bw, true);

        // Login
        toScreenAndLog("Login als User " + propUser + "...", bw, false);
        ftp.login(propUser, propPass);
        if (ftp.lastCode() != FTP.CODE_LOGGEDIN) {
        	toScreenAndLog("", bw, true);
        	toScreenAndLog("Fehler: Die FTP- oder Proxy-Zugangsdaten sind nicht korrekt. Programmende.", bw, true);
        	bw.close();
            ftp.disconnect();
            System.exit(0);
        }
        toScreenAndLog("fertig.", bw, true);
    	
    }

    public void deleteDir(File dir) {
    	   File[] files = dir.listFiles();
    	   if (files != null) {
    	      for (int i = 0; i < files.length; i++) {
    	         if (files[i].isDirectory()) {
    	            deleteDir(files[i]); // Verzeichnis leeren und anschlieend lschen
    	         }
    	         else {
    	            files[i].delete(); // Datei lschen
    	         }
    	      }
    	      dir.delete(); // Ordner lschen
    	   }
    	}
 }
